[TOC]

# RESTinio benchmark

This is a benchmark test suite for
[RESTinio](https://bitbucket.org/sobjectizerteam/restinio-0.4) -- a c++ REST framework.
As a developers of RESTinio we want to understand what is the performance of
our framework and how new changes influence it and, of course it is
interesting to compare RESTinio with other frameworks.

## Abstract

Assuming you have a piece of software that does REST API
and you using one of C++ REST frameworks providing request handler for it,
so what is the cost of the framework itself?
And that is the kind of question I am trying to answer with this benchmark.

Frameworks to test:

* [Beast](https://github.com/boostorg/beast) v123 (2017.10.15);
* [Cpprestsdk](https://github.com/Microsoft/cpprestsdk) commit `9747bc3b13581fc4462aabf96f484ca877ad6fd4` (2017.09.03).
	Included for comparison but not competing (as on linux it gives a poor performance);
* [Restbed](https://github.com/Corvusoft/restbed) 4.6;
* [Pistache](https://github.com/oktal/pistache) commit `b66415aaa0dc443b7874acbbb488058b34c2dc01` (2017.07.30);
* [Poco](https://pocoproject.org) 1.8.0.1 (2017.11.11).
* [CROW](https://github.com/ipkn/crow.git) `7f3f72441c242836d00ab2109a83baef89c08aaa` (2017.10.21).
* [RESTinio](https://bitbucket.org/sobjectizerteam/restinio-0.4) v.0.4.0.

More frameworks may be added later.

All benches are compared to a very straight ad-hock server implementation
with [ASIO](https://github.com/chriskohlhoff/asio.git).

## Choosing a starting point

Well, there are too many conditions that benchmarks depend on:
hardware, OS, compiler and compiler options,
and having raw numbers gives you not to much information.
So before starting with any framework bench,
lets implement a very quick server that pretends to be an http-server
and serves all request with the same response:
```
HTTP/1.1 200 OK
Connection: keep-alive
Content-Length: 12
Server: Pure asio bench server
Content-Type: text/plain; charset=utf-8

Hello world!
```

We've implemented such server using only [ASIO](https://github.com/chriskohlhoff/asio.git).
It receives connections and starts reading data until it finds "\r\n\r\n" sequence,
treats it as an end of request and sends prepared response back and then starts
reading again. No parsing, no handling, error check, just catching "\r\n\r\n" sequence
and serving hardcoded response.
And that logic is ok for GET requests of a benchmark load profile.
So the performance of that server I assume to be a pivot point for estimating
the performance of other frameworks on a single thread case.
See full implementation [here](./dev/pure_asio/benches/single_handler/main.cpp).

## Note on thread pool

Pure asio bench implementation and all framework-base implementations
(except cpprestsdk) have an option to set the number of threads to run on
(for asio or other io implementations).
So performance of cpprestsdk is measured and compared only for a single thread case.

## Benchmark tool and parameters

For performing tests a [wrk](https://github.com/wg/wrk) util was used
for generating a traffic of incoming requests.

Arguments for test were the following:
```
wrk -t T -c C -d 30 http://127.0.0.1:8080/
```
Where `t` defines the number of threads used by wkr, and `c` the number of running connections.
And for duration (`-d 30`) we decided to be fixed to 30s.

Each benchmark (except cpprestsdk) has a parameter `n` --
the number of threads used by http server for handling requests.

So the set of parameters can be describe with the follofing tupple: `(n, t, c)`;

Based on the hardware used for benchmarking we've picked a set of
reasonable parameters to test the frameworks:
```
n=1 t=1 c=8,16,32,64,128,256,512
n=1 t=2 c=16,32,64,128,256,512,1024
n=2 t=2 c=16,32,64,128,256,512,1024
n=2 t=4 c=32,64,128,256,512,1024,2048
n=4 t=4 c=32,64,128,256,512,1024,2048
```

And another set of parameter to see how each implementation scales:
```
n=1 t=4 c=256,d=90
n=2 t=4 c=256,d=90
n=3 t=4 c=256,d=90
n=4 t=4 c=256,d=90
```

If you you have a judgment on better parameters, please,
let us know (via bitbucket or by mail: ngrodzitski (at) stiffstream.com).

## Environment

The following environment was used for benchmarks:

* CPU: 8x Intel(R) Core(TM) i7-6700K CPU @ 4.00GHz;
* Memory: 16343MB;
* Operating System: Ubuntu 16.04.2 LTS.
* Compiler: gcc version 7.1.0 (Ubuntu 7.1.0-5ubuntu2~16.04)

## Running benchmark

For a given params each framework were tested in following way:

* test run for warmup, results are skipped (duration: 10s);
* wait for 1 second;
* run 3 times;
* take median result.

## RESULTS

Well, benchmarks with very simple logic (like in this case) cannot be considered a firm base
for deciding what framework to use in you project.
But still some conclusions can be made.

### Various parameters

![n=1 t=1 c=8,16,32,64,128,256,512](./graphs/n1t1.png)

![n=1 t=2 c=16,32,64,128,256,512,1024](./graphs/n1t2.png)

![n=2 t=2 c=16,32,64,128,256,512,1024](./graphs/n2t2.png)

![n=2 t=4 c=32,64,128,256,512,1024,2048](./graphs/n2t4.png)

![n=4 t=4 c=32,64,128,256,512,1024,2048](./graphs/n4t4.png)

See full results [here](https://docs.google.com/spreadsheets/d/e/2PACX-1vT0-bRITF8FqiGaC6vIRaQEMZwOu2ieu2gq0VuJvhnDQOInvYeYpPQAPKk-8SFpeE8HHChYrzOHWm3U/pubhtml).

**RESTinio implementations**

| Params              | Pure ASIO impl | RESTinio  | Relative performance |
|---------------------|----------------|-----------|----------------------|
| n=1 t=1 c=8 d=30    | 351380.58      | 175445.36 | 49.93%               |
| n=1 t=1 c=16 d=30   | 361579.35      | 175060.38 | 48.42%               |
| n=1 t=1 c=32 d=30   | 367785.93      | 178959.2  | 48.66%               |
| n=1 t=1 c=64 d=30   | 372382.68      | 177657.03 | 47.71%               |
| n=1 t=1 c=128 d=30  | 349201.62      | 182563.03 | 52.28%               |
| n=1 t=1 c=256 d=30  | 377925.54      | 181015.56 | 47.90%               |
| n=1 t=1 c=512 d=30  | 363060.15      | 171502.99 | 47.24%               |
| n=1 t=2 c=16 d=30   | 351380.58      | 174026.19 | 49.53%               |
| n=1 t=2 c=32 d=30   | 361579.35      | 174680.76 | 48.31%               |
| n=1 t=2 c=64 d=30   | 367785.93      | 177906.48 | 48.37%               |
| n=1 t=2 c=128 d=30  | 372382.68      | 181614.43 | 48.77%               |
| n=1 t=2 c=256 d=30  | 349201.62      | 175681.91 | 50.31%               |
| n=1 t=2 c=512 d=30  | 377925.54      | 167787.27 | 44.40%               |
| n=1 t=2 c=1024 d=30 | 363060.15      | 143724.15 | 39.59%               |
| n=2 t=2 c=16 d=30   | 653434.25      | 266198.97 | 40.74%               |
| n=2 t=2 c=32 d=30   | 669500.07      | 270346.4  | 40.38%               |
| n=2 t=2 c=64 d=30   | 680652.85      | 276916.26 | 40.68%               |
| n=2 t=2 c=128 d=30  | 682598.76      | 277661.61 | 40.68%               |
| n=2 t=2 c=256 d=30  | 667997.08      | 280576.45 | 42.00%               |
| n=2 t=2 c=512 d=30  | 653264.22      | 265045.3  | 40.57%               |
| n=2 t=2 c=1024 d=30 | 509655.35      | 240946.79 | 47.28%               |
| n=2 t=4 c=32 d=30   | 627274.38      | 258607.18 | 41.23%               |
| n=2 t=4 c=64 d=30   | 650457.59      | 271521.33 | 41.74%               |
| n=2 t=4 c=128 d=30  | 655013.98      | 270551.32 | 41.30%               |
| n=2 t=4 c=256 d=30  | 663064.18      | 264964.51 | 39.96%               |
| n=2 t=4 c=512 d=30  | 659984.89      | 266792.08 | 40.42%               |
| n=2 t=4 c=1024 d=30 | 524569         | 236217.29 | 45.03%               |
| n=2 t=4 c=2048 d=30 | 436297         | 233970.55 | 53.63%               |
| n=4 t=4 c=32 d=30   | 857288.76      | 385826.43 | 45.01%               |
| n=4 t=4 c=64 d=30   | 840225.67      | 388752.57 | 46.27%               |
| n=4 t=4 c=128 d=30  | 850066.22      | 388322.23 | 45.68%               |
| n=4 t=4 c=256 d=30  | 835288.07      | 405742.3  | 48.58%               |
| n=4 t=4 c=512 d=30  | 827101.33      | 396141.33 | 47.90%               |
| n=4 t=4 c=1024 d=30 | 694056.76      | 369585.42 | 53.25%               |
| n=4 t=4 c=2048 d=30 | 596781.16      | 358373.39 | 60.05%               |

### Scale benchmarks

Also we wondered how various implementations would scale.
So we fix wrk parameters to: `-t 4 -c 256 -d=90`
and test all competitors using 1, 2, 3 and 4 threads:
```
n=1 t=4 c=256
n=2 t=4 c=256
n=3 t=4 c=256
n=4 t=4 c=256
```
![scaling](./graphs/scaling.png)

See full results [here](https://docs.google.com/spreadsheets/d/e/2PACX-1vT0-bRITF8FqiGaC6vIRaQEMZwOu2ieu2gq0VuJvhnDQOInvYeYpPQAPKk-8SFpeE8HHChYrzOHWm3U/pubhtml).

**RESTinio implementation**

| Params             | Pure ASIO impl | RESTinio  | Relative performance |
|--------------------|----------------|-----------|----------------------|
| n=1 t=4 c=256 d=90 | 445260.32      | 185326.21 | 41.64%               |
| n=2 t=4 c=256 d=90 | 675205.30      | 276874.64 | 42.16%               |
| n=3 t=4 c=256 d=90 | 782043.77      | 359713.08 | 46.12%               |
| n=4 t=4 c=256 d=90 | 836979.59      | 396541.90 | 48.95%               |

## Build

For building benchmarks cmake is used. There are separate cmake scripts
for each framework benchmark, and it is a modification of the original cmake script
that comes with the framework.

First you need to obtain sources of all frameworks and their dependencies.
Ones it is done you can use a helper script (`build_benches.sh`)
that will build all the benchmarks and put all the binaries in *_bench_run/bin* dir.

Note that for frameworks that are as shared libraries you need to
point an library path to run their benchmarks, for example the relative path to
bin directory with benchmarks executables:
```
export LD_LIBRARY_PATH=../lib
```

For sample instructions on how to build benchmarks refer to the following two section.

### Obtain and build from archive

```
wget https://bitbucket.org/sobjectizerteam/restinio-benchmark/downloads/restinio-benchmark-0.4.0.tar.bz2
tar xjvf restinio-benchmark-0.4.0.tar.bz2
cd restinio-benchmark-0.4.0/dev
./build_benches.sh
export LD_LIBRARY_PATH=$LD_LIBRARY_PATH:`pwd`/_bench_run/lib
cd _bench_run/bin
```

### Obtain and build from repository

To build benchmarks from repository, you sould have Ruby and Mxx_ru (gem) installed.

```
hg clone https://bitbucket.org/sobjectizerteam/restinio-benchmark
cd restinio-benchmark
hg up v.0.4.0
mxxruexternals
cd dev
./build_benches.sh
export LD_LIBRARY_PATH=$LD_LIBRARY_PATH:`pwd`/_bench_run/lib
cd _bench_run/bin
```
