#include <cpprest/http_listener.h>
#include <cpprest/uri.h>
#include <cpprest/asyncrt_utils.h>
#include <cpprest/json.h>
#include <cpprest/filestream.h>
#include <cpprest/containerstream.h>
#include <cpprest/producerconsumerstream.h>

#include <app_args.hpp>

class handler
{
  public:
    handler( web::http::uri uri )
      :  m_listener{ std::move( uri ) }
    {
      m_listener.support(
        web::http::methods::GET,
        []( web::http::http_request message ){

          message.reply( web::http::status_codes::OK, "Hello world!" )
            .then( [](pplx::task<void> t)
            {
              try{
                t.get();
              }
              catch(...){
                //
              }
            } );
        } );
    }

    pplx::task<void> open(){ return m_listener.open(); }
    pplx::task<void> close(){ return m_listener.close(); }

  private:
    web::http::experimental::listener::http_listener m_listener;
};

int main( int argc, const char *argv[] )
{
  try
  {
    const app_args_t args{ argc, argv };

    if( !args.m_help )
    {
      utility::string_t address{ "http://localhost:" };
      address.append( std::to_string( args.m_port ) );

      std::cout << "pool size cannot be set for cpprestsdk" << std::endl;

      handler h( address );
      h.open().wait();
      std::cout << "Listening for requests at: " << address << std::endl;

      std::cout << "Press ENTER to exit." << std::endl;
      std::string line;
      std::getline( std::cin, line );

      h.close().wait();
    }
  }
  catch( const std::exception & ex )
  {
    std::cerr << "Error: " << ex.what() << std::endl;
    return 1;
  }

  return 0;
}
