/*
   Mathieu Stefani, 13 février 2016

   Example of an hello world server

   Adopted for benchmark with other frameworks.
*/

#include <pistache/endpoint.h>

#include <app_args.hpp>

using namespace Pistache;
// using namespace Net;

class HelloHandler : public Http::Handler {
public:

    HTTP_PROTOTYPE(HelloHandler)

    void onRequest(const Http::Request& request, Http::ResponseWriter response) {
        auto m = MIME( Text, Plain );
        m.setParam( "charset", "utf-8");

        response.headers()
            .add< Http::Header::Server >("Pistache benchmark")
            .add< Http::Header::ContentType>( m );

        response.send( Http::Code::Ok, "Hello World!" );
    }
};

int main( int argc, const char * argv[] ) {
    try
    {
        const app_args_t args{ argc, argv };

        if( !args.m_help )
        {
            if( 1 > args.m_pool_size )
                throw std::runtime_error{ "invalid thread_pool_size" };

            std::cout << "pool size: " << args.m_pool_size << std::endl;

            Address addr(Ipv4::any(), Port(args.m_port));
            auto opts = Http::Endpoint::options()
                .threads( args.m_pool_size );

            Http::Endpoint server(addr);
            server.init(opts);
            server.setHandler(Http::make_handler<HelloHandler>());
            server.serve();

            server.shutdown();
        }
    }
    catch( const std::exception & ex )
    {
        std::cerr << "Error: " << ex.what() << std::endl;
        return 1;
    }

    return 0;
}
