//
// adapted from Poco/Net/samples
//

#include "Poco/Net/HTTPServer.h"
#include "Poco/Net/HTTPRequestHandler.h"
#include "Poco/Net/HTTPRequestHandlerFactory.h"
#include "Poco/Net/HTTPServerParams.h"
#include "Poco/Net/HTTPServerRequest.h"
#include "Poco/Net/HTTPServerResponse.h"
#include "Poco/Net/HTTPServerParams.h"
#include "Poco/Net/ServerSocket.h"
#include "Poco/Timestamp.h"
#include "Poco/DateTimeFormatter.h"
#include "Poco/DateTimeFormat.h"
#include "Poco/Exception.h"
#include "Poco/ThreadPool.h"
#include "Poco/Util/ServerApplication.h"
#include "Poco/Util/Option.h"
#include "Poco/Util/OptionSet.h"
#include "Poco/Util/HelpFormatter.h"
#include <iostream>

#include <app_args.hpp>

using Poco::Net::ServerSocket;
using Poco::Net::HTTPRequestHandler;
using Poco::Net::HTTPRequestHandlerFactory;
using Poco::Net::HTTPServer;
using Poco::Net::HTTPServerRequest;
using Poco::Net::HTTPServerResponse;
using Poco::Net::HTTPServerParams;
using Poco::Timestamp;
using Poco::DateTimeFormatter;
using Poco::DateTimeFormat;
using Poco::ThreadPool;
using Poco::Util::ServerApplication;
using Poco::Util::Application;
using Poco::Util::Option;
using Poco::Util::OptionSet;
using Poco::Util::HelpFormatter;

const std::string RESP_BODY{ "Hello world!" };

class BenchRequestHandler: public Poco::Net::HTTPRequestHandler
{
public:
	BenchRequestHandler()
	{
	}

	void handleRequest(HTTPServerRequest& request, HTTPServerResponse& response)
	{
		response.setContentType("text/plain; charset=utf-8");
		response.set("Server", "Poco" );
		response.setKeepAlive( true );

		response.setContentLength(RESP_BODY.size());
		std::ostream& ostr = response.send();
		ostr << RESP_BODY;

		// This happen to give less performance:
		// response.sendBuffer(RESP_BODY.size(), RESP_BODY.size());
	}
};

class BenchRequestHandlerFactory: public HTTPRequestHandlerFactory
{
public:
	BenchRequestHandlerFactory()
	{
	}

	HTTPRequestHandler* createRequestHandler(const HTTPServerRequest& request)
	{
		return new BenchRequestHandler{};
	}
};

int main(int argc, const char* argv[])
{
	try
	{
		const app_args_t args{ argc, argv };

		if( !args.m_help )
		{
			std::cout << "pool size: " << args.m_pool_size << std::endl;

			if( 0 < args.m_pool_size )
			{
				ThreadPool::defaultPool().addCapacity(args.m_pool_size);
				HTTPServerParams* pParams = new HTTPServerParams;
				pParams->setMaxQueued(100);
				pParams->setMaxThreads(args.m_pool_size);

				// set-up a server socket
				ServerSocket svs(args.m_port);
				// set-up a HTTPServer instance
				HTTPServer srv(new BenchRequestHandlerFactory{}, svs, pParams);
				// start the HTTPServer
				srv.start();

				std::cout << "Press ENTER to exit." << std::endl;
				std::string line;
				std::getline( std::cin, line );

				// Stop the HTTPServer
				srv.stop();
			}
			else
			{
				throw std::runtime_error{ "invalid pool size" };
			}
		}
	}
	catch( const std::exception & ex )
	{
		std::cerr << "Error: " << ex.what() << std::endl;
		return 1;
	}
}
