#include <string>
#include <memory>
#include <cstdlib>
#include <stdexcept>

#include <restbed>

#include <app_args.hpp>

using namespace std;
using namespace restbed;

void get_method_handler( const shared_ptr< Session > session )
{
    const auto request = session->get_request( );

    session->yield( OK, "Hello world!", { { "Content-Length", "12" }, { "Connection", "keep-alive" } } );
}

void error_handler( const int status_code, const exception error, const shared_ptr< Session > session )
{
    if ( session not_eq nullptr and session->is_open( ) )
    {
        string message = error.what( );
        message.push_back( '\n' );

        //we must leave the socket intact on error,
        //otherwise follow up messages will be lost.
        session->yield( status_code, message, { { "Content-Length", std::to_string( message.length( ) ) }, { "Connection", "keep-alive" } } );
    }
}

int main( const int argc, const char** argv )
{
    try
    {
        const app_args_t args{ argc, argv };

        if( !args.m_help )
        {
            if( 1 > args.m_pool_size )
                throw std::runtime_error{ "invalid thread_pool_size" };

            std::cout << "pool size: " << args.m_pool_size << std::endl;

            auto resource = make_shared< Resource >( );
            resource->set_path( "/" );
            resource->set_method_handler( "GET", get_method_handler );

            auto settings = make_shared< Settings >( );
            settings->set_port( args.m_port );
            settings->set_worker_limit( args.m_pool_size );

            Service service;
            service.publish( resource );
            service.set_error_handler( error_handler );
            service.start( settings );
        }
    }
    catch( const std::exception & ex )
    {
        std::cerr << "Error: " << ex.what() << std::endl;
        return 1;
    }

    return EXIT_SUCCESS;
}
